import ij.*;
import ij.process.*;
import ij.gui.*;
import java.awt.*;
import ij.plugin.*;
import ij.plugin.frame.*;
import ij.plugin.filter.PlugInFilter;
import ij.*;
import ij.io.*;
import ij.plugin.*;
import ij.plugin.filter.*;

import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;
import java.util.zip.*;
import java.io.*;
import javax.swing.*;
import javax.swing.event.*;
import java.lang.InterruptedException;
import java.lang.System;
import java.lang.Math.*;

///Plugin class to load up a Thunderstorm run
///
///@ingroup gPlugin
public class  CoxGroupThunderstormLoader implements PlugIn {
	static 
	{
		new CoxGroupCheckForUpdates().run("boot");
	}

	private void close(InputStream i)
	{
		//Not sure what to do here...
		if(i != null)
		{
			try{
				i.close();
			}
			catch(IOException ioe)
			{
				Toolkit.getDefaultToolkit().beep();
				ij.IJ.showStatus("Error closing file: " + ioe.getMessage());
			}
		}
	}

	public void run(String arg) {
		InputStream in = null;
		String name;
		double ini_FWHM=100., ini_reconstruction_pixel_size=30., ini_sigma_filter=100;
		boolean show_control_panel=true;
		boolean do_filtering=true;

		//Make a fake pixel size because we need one: the reconstruction works in
		//pixel coordinates, because 3B unwisely(?) outputted only pixel coordinates
		//and no nanometers per pixel.  ThunderStorm (and QP) both outut locations
		//in nm, (QP also emits pixels), so we have to invent nm/pixel to convert to
		//hypothetical pixel coordinates to make the reconstruction work.
		final double pixel_size_in_nm_ = 200; 
		try{

			if(arg.length() != 0 && arg.substring(0,1).equals("!"))
			{
				in = getClass().getClassLoader().getResourceAsStream(arg.substring(1));
				name = arg.substring(1);
			}
			else
			{
				OpenDialog o = new OpenDialog("Open Thunderstorm data...", null);
				name = o.getFileName();
				
				try{
					in = new FileInputStream(o.getDirectory() + o.getFileName());
				}
				catch(java.io.FileNotFoundException ferr)
				{
					Toolkit.getDefaultToolkit().beep();
					ij.IJ.showStatus("Error opening file: " + ferr.getMessage());
					return;
				}


				//We have this rather than simply choosing defaults so that
				//a headless script can reconstruct a file and export it.
				GenericDialog g = new GenericDialog("Reconstruction Parameters");
				g.addNumericField("FWHM (initial) ", ini_FWHM, 0, 5, "nm");
				g.addNumericField("Reconstruction size (initial)", ini_reconstruction_pixel_size, 0, 5, "nm");
				g.addNumericField("Small localisation filter (initial)", ini_sigma_filter, 0, 5, "nm");
				g.addCheckbox("Show control panel", show_control_panel);
				
				((Checkbox)g.getCheckboxes().get(0)).hide();
				
				g.showDialog();

				ini_FWHM = g.getNextNumber();
				ini_reconstruction_pixel_size=g.getNextNumber();
				ini_sigma_filter = g.getNextNumber();
				
				show_control_panel=g.getNextBoolean();
			}

			//Yay @ cargoculting
			InputStreamReader d = new InputStreamReader(in);
			BufferedReader r = new BufferedReader(d);

			String line;
			final ArrayList<Spot> spots = new ArrayList<Spot>();
			Rectangle roi=null;


			int line_number=0;
			double max_val = 0; //Assume square image: almost all scientific imagers are close enough to square
			
			//Read the header and infer the format.
			int x_index=-1;
			int y_index=-1;
			int sigma_index=-1;

			line = r.readLine();
			if(line == null)
				throw new IOException("corrupt file (no header)");

			//Try assuming it's a TSV (what ThunderStorm reasonably calls an XLS),
			//and if the results don't look right, try assuming it's a CSV.
			boolean tsv=true;
			String tokens[] = line.split("\\t");
			if(tokens.length == 1)
			{
				tokens = line.split(",");
				tsv=false;
			}

			System.out.println("file type = " + tsv);
			System.out.println("line length = " + tokens.length);
			
			//Now find the relevant parts of the header.
			for(int i=0; i < tokens.length; i++)
			{
				System.out.println("Header token " + i + " = " + tokens[i]);

				if(tokens[i].equals("\"x [nm]\""))
					x_index = i;
				else if(tokens[i].equals("\"y [nm]\""))
					y_index = i;
				else if(tokens[i].equals("\"sigma [nm]\""))
					sigma_index = i;
			}

			System.out.println("x index = " + x_index);
			System.out.println("y index = " + y_index);
			System.out.println("s index = " + sigma_index);

			if(x_index == -1 || y_index == -1 || sigma_index == -1)
				throw new IOException("corrupt file (bad header)");



			for(  ; (line = r.readLine()) != null; line_number++)
			{
				if(tsv)
					tokens = line.split("\\t");
				else
					tokens = line.split(",");
				
				Spot s = new Spot();
				try
				{
					s.x = Double.parseDouble(tokens[x_index]) / pixel_size_in_nm_;
					s.y = Double.parseDouble(tokens[y_index]) / pixel_size_in_nm_;
					s.sigma = Double.parseDouble(tokens[sigma_index]) / pixel_size_in_nm_;


					if(s.x > max_val)
						max_val = s.x;
					if(s.y > max_val)
						max_val = s.y;
				}
				catch(NumberFormatException nerr)
				{
					throw new IOException("corrupt file (bad spot position)");
				}

				spots.add(s);
			}

			roi = new Rectangle(0, 0, (int)Math.ceil(max_val), (int)Math.ceil(max_val));


			if(show_control_panel)
			{
				final String fname = name;
				final Rectangle roi_ = roi;
				final int its = 0;
				final double ini_FWHM_ = ini_FWHM;
				final double ini_reconstruction_pixel_size_ = ini_reconstruction_pixel_size;
				final String msg = "Using " + fname + ": " + Integer.toString(spots.size()) + " localisations .";
				final double ini_sigma_filter_ = ini_sigma_filter;
				SwingUtilities.invokeLater(
				new Runnable() {
						public void run() {
								CoxGroupReconstructionViewer e = new CoxGroupReconstructionViewer(roi_, pixel_size_in_nm_, fname, null, true);
								e.append(spots, 100000); // 100000 == frames, to stop "Export" button from being red as a warning. This qualifies as lolhax
								e.send_update_canvas_event();
								e.send_status_text_message(msg);
								e.set_reconstructed_pixel_size(ini_reconstruction_pixel_size_);
								e.set_reconstruction_blur_fwhm(ini_FWHM_);
								e.set_sigma_filter(ini_sigma_filter_);
								e.send_update_canvas_event();
						}

				}
				);
			}
			else
			{
				double zoom=pixel_size_in_nm_ / ini_reconstruction_pixel_size; 
				ImagePlus   linear_reconstruction;
				linear_reconstruction = new ImagePlus();
				linear_reconstruction.setProcessor(Reconstruction.reconstruct(roi, zoom, ini_FWHM, pixel_size_in_nm_, spots, ini_sigma_filter));
				linear_reconstruction.show();
			}




		}
		catch(IOException another_ioe){
			Toolkit.getDefaultToolkit().beep();
			ij.IJ.showStatus("Error reading data: " + another_ioe.getMessage());
		}
		finally{
			close(in);
		}

	}



}
